"""
Simulation
====================

This module of Spirit is used to run and monitor iterative calculation methods.

If many iterations are called individually, one should use the single shot simulation functionality.
It avoids the allocations etc. involved when a simulation is started and ended and behaves like a
regular simulation, except that the iterations have to be triggered manually.

Note that the VP and LBFGS Solvers are only meant for direct minimization and not for dynamics.
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

import threading


###     We use a thread for PlayPause, so that KeyboardInterrupt can be forwarded to the CDLL call
###     We might want to think about using PyDLL and about a signal handler in the core library
###     see here: http://stackoverflow.com/questions/14271697/ctrlc-doesnt-interrupt-call-to-shared-library-using-ctypes-in-python


SOLVER_VP = 0
"""Verlet-like velocity projection method."""

SOLVER_SIB = 1
"""Semi-implicit midpoint method B."""

SOLVER_DEPONDT = 2
"""Depondt's Heun-like method."""

SOLVER_HEUN = 3
"""Heun's method."""

SOLVER_RK4 = 4
"""4th order Runge-Kutta method."""

SOLVER_LBFGS_OSO = 5
"""Limited memory Broyden-Fletcher-Goldfarb-Shannon, using exponential transforms."""

SOLVER_LBFGS_Atlas = 6
"""Limited memory Broyden-Fletcher-Goldfarb-Shannon, using stereographic transforms"""

SOLVER_VP_OSO = 7
"""Verlet-like velocity projection method, using exponential transforms."""


METHOD_MC = 0
"""Monte Carlo.

Standard implementation.
"""

METHOD_LLG = 1
"""Landau-Lifshitz-Gilbert.

Can be either a dynamical simulation or an energy minimisation.
Note: the VP solver can *only* minimise.
"""

METHOD_GNEB = 2
"""Geodesic nudged elastic band.

Runs on the entire chain.

As this is a minimisation method, the dynamical solvers perform worse
than those designed for minimisation.
"""

METHOD_MMF = 3
"""Minimum mode following.

As this is a minimisation method, the dynamical solvers perform worse
than those designed for minimisation.
"""

METHOD_EMA = 4
"""Eigenmode analysis.

Applies eigenmodes to the spins of a system.
Depending on parameters, this can be used to calculate the change of a
spin configuration through such a mode or to get a "dynamical" chain
of images corresponding to the movement of the system under the mode.
"""


class simulation_run_info(ctypes.Structure):
    """Contains basic information about a simulation run."""

    _fields_ = [
        ("total_iterations", ctypes.c_int),
        ("total_walltime", ctypes.c_int),
        ("total_ips", ctypes.c_float),
        ("max_torque", ctypes.c_float),
        ("_n_history_iteration", ctypes.c_int),
        ("_history_iteration", ctypes.POINTER(ctypes.c_int)),
        ("_n_history_max_torque", ctypes.c_int),
        ("_history_max_torque", ctypes.POINTER(ctypes.c_float)),
        ("_n_history_energy", ctypes.c_int),
        ("_history_energy", ctypes.POINTER(ctypes.c_float)),
    ]

    def history_iteration(self):
        return [self._history_iteration[i] for i in range(self._n_history_iteration)]

    def history_max_torque(self):
        return [self._history_max_torque[i] for i in range(self._n_history_max_torque)]

    def history_energy(self):
        return [self._history_energy[i] for i in range(self._n_history_energy)]

    def __del__(self):
        _spirit.free_run_info(self)


### ----- Start methods
### MC
_MC_Start = _spirit.Simulation_MC_Start
_MC_Start.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_bool,
    ctypes.POINTER(simulation_run_info),
    ctypes.c_int,
    ctypes.c_int,
]
_MC_Start.restype = None
### LLG
_LLG_Start = _spirit.Simulation_LLG_Start
_LLG_Start.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_bool,
    ctypes.POINTER(simulation_run_info),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Start.restype = None
### GNEB
_GNEB_Start = _spirit.Simulation_GNEB_Start
_GNEB_Start.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_bool,
    ctypes.POINTER(simulation_run_info),
    ctypes.c_int,
]
_GNEB_Start.restype = None
### MMF
_MMF_Start = _spirit.Simulation_MMF_Start
_MMF_Start.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_bool,
    ctypes.POINTER(simulation_run_info),
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Start.restype = None
### EMA
_EMA_Start = _spirit.Simulation_EMA_Start
_EMA_Start.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_bool,
    ctypes.POINTER(simulation_run_info),
    ctypes.c_int,
    ctypes.c_int,
]
_EMA_Start.restype = None


### ----- Wrapper
def start(
    p_state,
    method_type,
    solver_type=None,
    n_iterations=-1,
    n_iterations_log=-1,
    single_shot=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Start any kind of iterative calculation method.

    - `method_type`: one of the integers defined above
    - `solver_type`: only used for LLG, GNEB and MMF methods (default: None)
    - `n_iterations`: the maximum number of iterations that will be performed (default: take from parameters)
    - `n_iterations_log`: the number of iterations after which to log the status and write output (default: take from parameters)
    - `single_shot`: if set to `True`, iterations have to be triggered individually
    - `idx_image`: the image on which to run the calculation (default: active image). Not used for GNEB

    returns a `simulation_run_info` object.
    """

    info = simulation_run_info()

    if method_type == METHOD_MC:
        spiritlib.wrap_function(
            _MC_Start,
            [
                ctypes.c_void_p(p_state),
                ctypes.c_int(n_iterations),
                ctypes.c_int(n_iterations_log),
                ctypes.c_bool(single_shot),
                ctypes.pointer(info),
                ctypes.c_int(idx_image),
                ctypes.c_int(idx_chain),
            ],
        )
    elif method_type == METHOD_LLG:
        spiritlib.wrap_function(
            _LLG_Start,
            [
                ctypes.c_void_p(p_state),
                ctypes.c_int(solver_type),
                ctypes.c_int(n_iterations),
                ctypes.c_int(n_iterations_log),
                ctypes.c_bool(single_shot),
                ctypes.pointer(info),
                ctypes.c_int(idx_image),
                ctypes.c_int(idx_chain),
            ],
        )
    elif method_type == METHOD_GNEB:
        spiritlib.wrap_function(
            _GNEB_Start,
            [
                ctypes.c_void_p(p_state),
                ctypes.c_int(solver_type),
                ctypes.c_int(n_iterations),
                ctypes.c_int(n_iterations_log),
                ctypes.c_bool(single_shot),
                ctypes.pointer(info),
                ctypes.c_int(idx_chain),
            ],
        )
    elif method_type == METHOD_MMF:
        spiritlib.wrap_function(
            _MMF_Start,
            [
                ctypes.c_void_p(p_state),
                ctypes.c_int(solver_type),
                ctypes.c_int(n_iterations),
                ctypes.c_int(n_iterations_log),
                ctypes.c_bool(single_shot),
                ctypes.pointer(info),
                ctypes.c_int(idx_image),
                ctypes.c_int(idx_chain),
            ],
        )
    elif method_type == METHOD_EMA:
        spiritlib.wrap_function(
            _EMA_Start,
            [
                ctypes.c_void_p(p_state),
                ctypes.c_int(n_iterations),
                ctypes.c_int(n_iterations_log),
                ctypes.c_bool(single_shot),
                ctypes.pointer(info),
                ctypes.c_int(idx_image),
                ctypes.c_int(idx_chain),
            ],
        )
    else:
        print("Invalid method_type passed to simulation.start...")

    return info
    # _Start(ctypes.c_void_p(p_state), ctypes.c_char_p(method_type),
    #            ctypes.c_char_p(solver_type), ctypes.c_int(n_iterations),
    #            ctypes.c_int(n_iterations_log), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))


_SingleShot = _spirit.Simulation_SingleShot
_SingleShot.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_SingleShot.restype = None


def single_shot(p_state, idx_image=-1, idx_chain=-1):
    """Perform a single iteration.

    In order to use this, a single shot simulation must be running on the corresponding image or chain.
    """
    spiritlib.wrap_function(
        _SingleShot,
        [ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)],
    )


_N_Shot = _spirit.Simulation_N_Shot
_N_Shot.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_N_Shot.restype = None


def n_shot(p_state, N, idx_image=-1, idx_chain=-1):
    """Perform a single iteration.

    In order to use this, a single shot simulation must be running on the corresponding image or chain.
    """
    spiritlib.wrap_function(
        _N_Shot,
        [
            ctypes.c_void_p(p_state),
            ctypes.c_int(N),
            ctypes.c_int(idx_image),
            ctypes.c_int(idx_chain),
        ],
    )


_Stop = _spirit.Simulation_Stop
_Stop.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Stop.restype = None


def stop(p_state, idx_image=-1, idx_chain=-1):
    """Stop the simulation running on an image or chain."""
    _Stop(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))


_Stop_All = _spirit.Simulation_Stop_All
_Stop_All.argtypes = [ctypes.c_void_p]
_Stop_All.restype = None


def stop_all(p_state):
    """Stop all simulations running anywhere."""
    _Stop_All(ctypes.c_void_p(p_state))


_Running_On_Image = _spirit.Simulation_Running_On_Image
_Running_On_Image.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Running_On_Image.restype = ctypes.c_bool


def running_on_image(p_state, idx_image=-1, idx_chain=-1):
    """Check if a simulation is running on a specific image."""
    return bool(
        _Running_On_Image(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_Running_On_Chain = _spirit.Simulation_Running_On_Chain
_Running_On_Chain.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Running_On_Chain.restype = ctypes.c_bool


def running_on_chain(p_state, idx_chain=-1):
    """Check if a simulation is running across a specific chain."""
    return bool(_Running_On_Chain(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)))


_Running_Anywhere_On_Chain = _spirit.Simulation_Running_Anywhere_On_Chain
_Running_Anywhere_On_Chain.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Running_Anywhere_On_Chain.restype = ctypes.c_bool


def running_anywhere_on_chain(p_state, idx_chain=-1):
    """Check if any simulation running on any image of - or the entire - chain."""
    return bool(
        _Running_Anywhere_On_Chain(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain))
    )


_Get_IterationsPerSecond = _spirit.Simulation_Get_IterationsPerSecond
_Get_IterationsPerSecond.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_IterationsPerSecond.restype = ctypes.c_float


def get_iterations_per_second(p_state, idx_image=-1, idx_chain=-1):
    """Returns the current estimation of the number of iterations per second."""
    return float(
        _Get_IterationsPerSecond(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_Get_MaxTorqueNorm = _spirit.Simulation_Get_MaxTorqueNorm
_Get_MaxTorqueNorm.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_MaxTorqueNorm.restype = ctypes.c_float


def get_max_torque_norm(p_state, idx_image=-1, idx_chain=-1):
    """Returns the current maximum norm of the torque acting on any spin."""
    return float(
        _Get_MaxTorqueNorm(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_Get_Time = _spirit.Simulation_Get_Time
_Get_Time.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Time.restype = ctypes.c_float


def get_time(p_state, idx_image=-1, idx_chain=-1):
    """If an LLG simulation is running returns the cumulatively summed time steps `dt`, otherwise returns 0"""
    return _Get_Time(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


_Get_Wall_Time = _spirit.Simulation_Get_Wall_Time
_Get_Wall_Time.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Wall_Time.restype = ctypes.c_int


def get_wall_time(p_state, idx_image=-1, idx_chain=-1):
    """Returns the current maximum norm of the torque acting on any spin."""
    return int(
        _Get_Wall_Time(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )



"""
I/O
====================

Read and write spin configurations, chains or eigenmodes.
Vectorfields are generally written in the `OOMMF vector field (OVF) file format <http://math.nist.gov/oommf/doc/userguide12a5/userguide/OVF_2.0_format.html>`_.

Note that, when reading an image or chain from file, the file will automatically be tested for an OVF header.
If it cannot be identified as OVF, it will be tried to be read as three plain text columns (Sx Sy Sz).

Note also, IO is still being re-written and only OVF will be supported as output format.
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### Output file formats
FILEFORMAT_OVF_BIN = 0
"""OVF binary format corresponding to the precision with which the code was compiled"""

FILEFORMAT_OVF_BIN4 = 1
"""OVF binary format with precision 4"""

FILEFORMAT_OVF_BIN8 = 2
"""OVF binary format with precision 8"""

FILEFORMAT_OVF_TEXT = 3
"""OVF text format"""

FILEFORMAT_OVF_CSV = 4
"""OVF text format with comma-separated columns"""

_N_Images_In_File = _spirit.IO_N_Images_In_File
_N_Images_In_File.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_N_Images_In_File.restype = ctypes.c_int


def n_images_in_file(p_state, filename, idx_image_inchain=-1, idx_chain=-1):
    """Returns the number of segments or images in a given file.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to check
    """
    return int(
        _N_Images_In_File(
            ctypes.c_void_p(p_state),
            ctypes.c_char_p(filename.encode("utf-8")),
            ctypes.c_int(idx_image_inchain),
            ctypes.c_int(idx_chain),
        )
    )


_Image_Read = _spirit.IO_Image_Read
_Image_Read.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_Image_Read.restype = None


def image_read(
    p_state, filename, idx_image_infile=0, idx_image_inchain=-1, idx_chain=-1
):
    """Attempt to read a spin configuration from a file into an image of the chain.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to read

    Keyword arguments:

    - `idx_image_infile`: the index of the image in the file which should be read in (default: 0)
    - `idx_image_inchain`: the index of the image in the chain into which the data should be read (default: active image)
    """
    _Image_Read(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(idx_image_infile),
        ctypes.c_int(idx_image_inchain),
        ctypes.c_int(idx_chain),
    )


_Image_Write = _spirit.IO_Image_Write
_Image_Write.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_Image_Write.restype = None


def image_write(
    p_state,
    filename,
    fileformat=FILEFORMAT_OVF_TEXT,
    comment="",
    idx_image=-1,
    idx_chain=-1,
):
    """Write an image of the chain to a file.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to write

    Keyword arguments:

    - `fileformat`: the format in which to write the data (default: OVF text)
    - `comment`: a comment string to be inserted in the header (default: empty)
    - `idx_image`: the index of the image to be written to the file (default: active image)
    """
    _Image_Write(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_char_p(comment.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Image_Append = _spirit.IO_Image_Append
_Image_Append.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_Image_Append.restype = None


def image_append(
    p_state,
    filename,
    fileformat=FILEFORMAT_OVF_TEXT,
    comment="",
    idx_image=-1,
    idx_chain=-1,
):
    """Append an image of the chain to a file, incrementing the segment count.

    If the file does not exist, it is created.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to append to

    Keyword arguments:

    - `fileformat`: the format in which to write the data (default: OVF text)
    - `comment`: a comment string to be inserted in the header (default: empty)
    - `idx_image`: the index of the image to be written to the file (default: active image)
    """
    _Image_Append(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Chain_Read = _spirit.IO_Chain_Read
_Chain_Read.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_Chain_Read.restype = None


def chain_read(
    p_state, filename, starting_image=0, ending_image=-1, insert_idx=-1, idx_chain=-1
):
    """Attempt to read a chain of images from a given file.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to read

    Keyword arguments:

    - `starting_image`: the index within the file at which to start reading (default: 0)
    - `ending_image`: the index within the file at which to stop reading (default: -1, meaning the entire file)
    - `insert_idx`: the index within the chain at which to start placing the images (default: active image)

    Images of the chain will be overwritten with what is read from the file.
    If the chain is not long enough for the number of images to be read, it is automatically set to the
    right length.
    """
    _Chain_Read(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(starting_image),
        ctypes.c_int(ending_image),
        ctypes.c_int(insert_idx),
        ctypes.c_int(idx_chain),
    )


_Chain_Write = _spirit.IO_Chain_Write
_Chain_Write.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_char_p,
    ctypes.c_int,
]
_Chain_Write.restype = None


def chain_write(
    p_state, filename, fileformat=FILEFORMAT_OVF_TEXT, comment="", idx_chain=-1
):
    """Write a chain of images to a file.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to write

    Keyword arguments:

    - `fileformat`: the format in which to write the data (default: OVF text)
    - `comment`: a comment string to be inserted in the header (default: empty)
    """
    _Chain_Write(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_char_p(comment.encode("utf-8")),
        ctypes.c_int(idx_chain),
    )


_Chain_Append = _spirit.IO_Chain_Append
_Chain_Append.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_char_p,
    ctypes.c_int,
]
_Chain_Append.restype = None


def chain_append(
    p_state, filename, fileformat=FILEFORMAT_OVF_TEXT, comment="", idx_chain=-1
):
    """Append a chain of images to a given file.

    If the file does not exist, it is created.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to append to

    Keyword arguments:

    - `fileformat`: the format in which to write the data (default: OVF text)
    - `comment`: a comment string to be inserted in the header (default: empty)
    """
    _Chain_Append(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_char_p(comment.encode("utf-8")),
        ctypes.c_int(idx_chain),
    )


_Eigenmodes_Read = _spirit.IO_Eigenmodes_Read
_Eigenmodes_Read.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_Eigenmodes_Read.restype = None


def eigenmodes_read(
    p_state,
    filename,
    fileformat=FILEFORMAT_OVF_TEXT,
    idx_image_inchain=-1,
    idx_chain=-1,
):
    """Read the vector fields from a file as a set of eigenmodes for the spin system."""
    _Eigenmodes_Read(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_int(idx_image_inchain),
        ctypes.c_int(idx_chain),
    )


_Eigenmodes_Write = _spirit.IO_Eigenmodes_Write
_Eigenmodes_Write.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_Eigenmodes_Write.restype = None


def eigenmodes_write(
    p_state,
    filename,
    fileformat=FILEFORMAT_OVF_TEXT,
    comment="",
    idx_image=-1,
    idx_chain=-1,
):
    """Write the eigenmodes of a spin system to file, if they have been already calculated."""
    _Eigenmodes_Write(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_char_p(comment.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Image_Write_Energy_per_Spin = _spirit.IO_Image_Write_Energy_per_Spin
_Image_Write_Energy_per_Spin.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_Image_Write_Energy_per_Spin.restype = None


def image_write_energy_per_spin(
    p_state, filename, fileformat=FILEFORMAT_OVF_TEXT, idx_image=-1, idx_chain=-1
):
    """Write the energies per spin to a file.

    Arguments:

    - `p_state`: state pointer
    - `filename`: the name of the file to append to

    Keyword arguments

    - `fileformat`: the format in which to write the data (default: OVF text)
    """
    _Image_Write_Energy_per_Spin(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_int(fileformat),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )



"""
Eigenmode analysis (EMA)
-------------------------------------------------------------

This method, if needed, calculates modes (they can also be read in from a file)
and perturbs the spin system periodically in the direction of the eigenmode.
"""

from spirit import spiritlib
import ctypes

# Load Library
_spirit = spiritlib.load_spirit_library()

_EMA_Clear_Modes = _spirit.Parameters_EMA_Clear_Modes
_EMA_Clear_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_EMA_Clear_Modes.restype = None


def clear_modes(p_state, idx_image=-1, idx_chain=-1):
    """Clears the modes."""
    _EMA_Clear_Modes(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


# ---------------------------------- Set ----------------------------------

_EMA_Set_N_Modes = _spirit.Parameters_EMA_Set_N_Modes
_EMA_Set_N_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_EMA_Set_N_Modes.restype = None


def set_n_modes(p_state, n_modes, idx_image=-1, idx_chain=-1):
    """Set the number of modes to calculate or use."""
    _EMA_Set_N_Modes(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_modes),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_EMA_Set_N_Mode_Follow = _spirit.Parameters_EMA_Set_N_Mode_Follow
_EMA_Set_N_Mode_Follow.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_EMA_Set_N_Mode_Follow.restype = None


def set_n_mode_follow(p_state, n_mode, idx_image=-1, idx_chain=-1):
    """Set the index of the mode to use."""
    _EMA_Set_N_Mode_Follow(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_mode),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_EMA_Set_Sparse = _spirit.Parameters_EMA_Set_Sparse
_EMA_Set_Sparse.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_EMA_Set_Sparse.restype = None


def set_sparse(p_state, sparse, idx_image=-1, idx_chain=-1):
    """Set wether to use sparse matrices."""
    _EMA_Set_Sparse(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(sparse),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


# ---------------------------------- Get ----------------------------------

_EMA_Get_N_Modes = _spirit.Parameters_EMA_Get_N_Modes
_EMA_Get_N_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_EMA_Get_N_Modes.restype = ctypes.c_int


def get_n_modes(p_state, idx_image=-1, idx_chain=-1):
    """Returns the number of modes to calculate or use."""
    return int(
        _EMA_Get_N_Modes(p_state, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    )


_EMA_Get_N_Mode_Follow = _spirit.Parameters_EMA_Get_N_Mode_Follow
_EMA_Get_N_Mode_Follow.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_EMA_Get_N_Mode_Follow.restype = ctypes.c_int


def get_n_mode_follow(p_state, idx_image=-1, idx_chain=-1):
    """Returns the index of the mode to use."""
    return int(
        _EMA_Get_N_Mode_Follow(
            p_state, ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_EMA_Get_Sparse = _spirit.Parameters_EMA_Get_Sparse
_EMA_Get_Sparse.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_EMA_Get_Sparse.restype = ctypes.c_bool


def get_sparse(p_state, sparse, idx_image=-1, idx_chain=-1):
    """Set wether to use sparse matrices."""
    return bool(
        _EMA_Get_Sparse(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )



"""
Quantities
====================
"""

from spirit import spiritlib
from spirit import system
from spirit import scalar
import ctypes

import numpy as np

### Load Library
_spirit = spiritlib.load_spirit_library()

_Get_Magnetization = _spirit.Quantity_Get_Magnetization
_Get_Magnetization.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Magnetization.restype = None


def get_magnetization(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the average magnetization of the system as
    an array of `shape(3)`.
    """
    magnetization = (3 * ctypes.c_float)()
    _Get_Magnetization(
        ctypes.c_void_p(p_state),
        magnetization,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return [float(i) for i in magnetization]


### Temporary info function for MMF
_Get_MinimumMode = _spirit.Quantity_Get_Grad_Force_MinimumMode
_Get_MinimumMode.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_MinimumMode.restype = None


def get_mmf_info(p_state, idx_image=-1, idx_chain=-1):
    """Returns a set of MMF information, meant mostly for testing or debugging.

    - `numpy.array_view` of `shape(NOS, 3)` of the energy gradient
    - the lowest eigenvalue
    - `numpy.array_view` of `shape(NOS, 3)` of the eigenmode
    - `numpy.array_view` of `shape(NOS, 3)` of the force
    """
    nos = system.get_nos(p_state, idx_image, idx_chain)

    ArrayType = ctypes.c_float * (3 * nos)

    MM = [] * (3 * nos)
    _MM = ArrayType(*MM)
    FF = [] * (3 * nos)
    _FF = ArrayType(*FF)
    GG = [] * (3 * nos)
    _GG = ArrayType(*FF)
    ev = [] * 1
    _eval = ArrayType(*ev)

    _Get_MinimumMode(p_state, _GG, _eval, _MM, _FF, idx_image, idx_chain)

    # array_pointer = ctypes.cast(_MM, ctypes.POINTER(ArrayType))
    # array = np.frombuffer(array_pointer.contents)

    MMM = np.array(_MM)
    FFF = np.array(_FF)
    GGG = np.array(_GG)
    array_view_mode = MMM.view()
    array_view_mode.shape = (nos, 3)
    array_view_force = FFF.view()
    array_view_force.shape = (nos, 3)
    array_view_grad = GGG.view()
    array_view_grad.shape = (nos, 3)

    return array_view_grad, _eval[0], array_view_mode, array_view_force


_Get_Topological_Charge = _spirit.Quantity_Get_Topological_Charge
_Get_Topological_Charge.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Topological_Charge.restype = ctypes.c_float


def get_topological_charge(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the total topological charge of 2D systems.

    Note that the charge can take unphysical non-integer values for open boundaries,
    because it is not well-defined in this case.

    Returns 0 for systems of other dimensionality.
    """
    return float(
        _Get_Topological_Charge(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_Get_Topological_Charge_Density = _spirit.Quantity_Get_Topological_Charge_Density
_Get_Topological_Charge_Density.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Topological_Charge_Density.restype = ctypes.c_int


def get_topological_charge_density(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the total topological charge of 2D systems.

    Note that the charge can take unphysical non-integer values for open boundaries,
    because it is not well-defined in this case.

    Returns 0 for systems of other dimensionality.
    """

    # Figure out the number of triangles
    num_triangles = _Get_Topological_Charge_Density(
        ctypes.c_void_p(p_state),
        None,
        None,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )

    # Allocate the required memory for indices_ptr and charge_density
    charge_density_ptr = (ctypes.c_float * num_triangles)()
    triangle_indices_ptr = (ctypes.c_int * (3 * num_triangles))()

    _Get_Topological_Charge_Density(
        ctypes.c_void_p(p_state),
        charge_density_ptr,
        triangle_indices_ptr,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )

    return [float(c) for c in charge_density_ptr], [
        [
            int(triangle_indices_ptr[3 * i]),
            int(triangle_indices_ptr[3 * i + 1]),
            int(triangle_indices_ptr[3 * i + 2]),
        ]
        for i in range(num_triangles)
    ]



"""
Chain
====================

Manipulate the chain of spin systems (also called images), e.g. add, remove or change active image.
Get information, such as number of images or energies and reaction coordinates.
"""

from spirit import spiritlib
from spirit import system
from spirit import parameters
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()


### Get Chain number of images
_Get_NOI = _spirit.Chain_Get_NOI
_Get_NOI.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Get_NOI.restype = ctypes.c_int


def get_noi(p_state, idx_chain=-1):
    """Get number of images (NOI) in the chain."""
    return int(_Get_NOI(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)))


### Switch active to next image of chain
_next_Image = _spirit.Chain_next_Image
_next_Image.argtypes = [ctypes.c_void_p, ctypes.c_int]
_next_Image.restype = ctypes.c_bool


def next_image(p_state, idx_chain=-1):
    """Switch the active image index to the next highest in the chain."""
    return bool(_next_Image(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)))


### Switch active to previous image of chain
_prev_Image = _spirit.Chain_prev_Image
_prev_Image.argtypes = [ctypes.c_void_p, ctypes.c_int]
_prev_Image.restype = ctypes.c_bool


def prev_image(p_state, idx_chain=-1):
    """Switch the active image index to the next lowest in the chain."""
    return bool(_prev_Image(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)))


### Switch active to specific image of chain
_Jump_To_Image = _spirit.Chain_Jump_To_Image
_Jump_To_Image.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Jump_To_Image.restype = ctypes.c_bool


def jump_to_image(p_state, idx_image=-1, idx_chain=-1):
    """Set the index of the active image in the chain."""
    return bool(
        _Jump_To_Image(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


### Set the number of images in the chain
_Set_Length = _spirit.Chain_Set_Length
_Set_Length.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Set_Length.restype = None


def set_length(p_state, n_images, idx_chain=-1):
    """Set the number of images (NOI) in the chain.

    If the chain is longer, the corresponding number of images is erased from the end.

    If the chain is shorter, the corresponding number of images is appended.

    Note that the active image might change.

    If no image is in the clipboard, no action is taken.
    """
    _Set_Length(
        ctypes.c_void_p(p_state), ctypes.c_int(n_images), ctypes.c_int(idx_chain)
    )


### Copy active image to clipboard
_Image_to_Clipboard = _spirit.Chain_Image_to_Clipboard
_Image_to_Clipboard.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Image_to_Clipboard.restype = None


def image_to_clipboard(p_state, idx_image=-1, idx_chain=-1):
    """Copies an image to the clipboard of Spirit. It can then be later e.g. inserted or appended."""
    _Image_to_Clipboard(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


### Replace active image in chain
_Replace_Image = _spirit.Chain_Replace_Image
_Replace_Image.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Replace_Image.restype = None


def replace_image(p_state, idx_image=-1, idx_chain=-1):
    """Replaces the image from the one in the clipboard.

    If no image is in the clipboard, no action is taken.
    """
    _Replace_Image(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


### Insert clipboard image before image in chain
_Insert_Image_Before = _spirit.Chain_Insert_Image_Before
_Insert_Image_Before.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Insert_Image_Before.restype = None


def insert_image_before(p_state, idx_image=-1, idx_chain=-1):
    """Inserts an image in front of the specified index.

    Note that the active image might change.

    If no image is in the clipboard, no action is taken.
    """
    _Insert_Image_Before(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


### Insert clipboard image after image in chain
_Insert_Image_After = _spirit.Chain_Insert_Image_After
_Insert_Image_After.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Insert_Image_After.restype = None


def insert_image_after(p_state, idx_image=-1, idx_chain=-1):
    """Inserts an image after the specified index.

    Note that the active image might change.

    If no image is in the clipboard, no action is taken.
    """
    _Insert_Image_After(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


### Insert clipboard image at end of chain
_Push_Back = _spirit.Chain_Push_Back
_Push_Back.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Push_Back.restype = None


def push_back(p_state, idx_chain=-1):
    """Appends an image to the chain.

    If no image is in the clipboard, no action is taken.
    """
    _Push_Back(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain))


### Delete active image
_Delete_Image = _spirit.Chain_Delete_Image
_Delete_Image.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Delete_Image.restype = ctypes.c_bool


def delete_image(p_state, idx_image=-1, idx_chain=-1):
    """Removes the specified image from the chain.

    Note that the active image might change.

    If it is the last remaining image in the chain, no action is taken.
    """
    return bool(
        _Delete_Image(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


### Delete image at end of chain
_Pop_Back = _spirit.Chain_Pop_Back
_Pop_Back.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Pop_Back.restype = ctypes.c_bool


def pop_back(p_state, idx_chain=-1):
    """Removes the last image from the chain.

    Note that the active image might change.

    If it is the last remaining image in the chain, no action is taken.
    """
    return bool(_Pop_Back(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)))


### Update the chain's data (interpolated energies etc.)
_Update_Data = _spirit.Chain_Update_Data
_Update_Data.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Update_Data.restype = None


def update_data(p_state, idx_chain=-1):
    """Updates various data of the chain, including:

    - Energies of images
    - Reaction coordinates of images
    - Interpolated energy and reaction coordinate values
    """
    _Update_Data(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain))


### Setup the chain's data arrays (when is this necessary?)
_Setup_Data = _spirit.Chain_Setup_Data
_Setup_Data.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Setup_Data.restype = None


def setup_data(p_state, idx_chain=-1):
    _Setup_Data(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain))


### Get Rx
_Get_Rx = _spirit.Chain_Get_Rx
_Get_Rx.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Rx.restype = None


def get_reaction_coordinate(p_state, idx_chain=-1):
    """Returns an array of `shape(NOI)` containing the reaction coordinates of the images."""
    noi = get_noi(p_state, idx_chain)
    Rx = (noi * ctypes.c_float)()
    _Get_Rx(ctypes.c_void_p(p_state), Rx, ctypes.c_int(idx_chain))
    return [x for x in Rx]


### Get Rx interpolated
_Get_Rx_Interpolated = _spirit.Chain_Get_Rx_Interpolated
_Get_Rx_Interpolated.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Rx_Interpolated.restype = None


def get_reaction_coordinate_interpolated(p_state, idx_chain=-1):
    """Returns an array containing the interpolated reaction coordinate values along the chain.

    The number of interpolated values between images can be set in the GNEB parameters.
    """
    noi = get_noi(p_state, idx_chain)
    n_interp = parameters.gneb.get_n_energy_interpolations(p_state, idx_chain)
    len_Rx = noi + (noi - 1) * n_interp
    Rx = (len_Rx * ctypes.c_float)()
    _Get_Rx_Interpolated(ctypes.c_void_p(p_state), Rx, ctypes.c_int(idx_chain))
    return [x for x in Rx]


### Get Energy
_Get_Energy = _spirit.Chain_Get_Energy
_Get_Energy.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Energy.restype = None


def get_energy(p_state, idx_chain=-1):
    """Returns an array of `shape(NOI)` containing the energies of the images."""
    noi = get_noi(p_state, idx_chain)
    Energy = (noi * ctypes.c_float)()
    _Get_Energy(ctypes.c_void_p(p_state), Energy, ctypes.c_int(idx_chain))
    return [E for E in Energy]


### Get Energy Interpolated
_Get_Energy_Interpolated = _spirit.Chain_Get_Energy_Interpolated
_Get_Energy_Interpolated.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Energy_Interpolated.restype = None


def get_energy_interpolated(p_state, idx_chain=-1):
    """Returns an array containing the interpolated energy values along the chain.

    The number of interpolated values between images can be set in the GNEB parameters.
    """
    noi = get_noi(p_state, idx_chain)
    n_interp = parameters.gneb.get_n_energy_interpolations(p_state, idx_chain)
    len_Energy = noi + (noi - 1) * n_interp
    Energy_interp = (len_Energy * ctypes.c_float)()
    _Get_Energy_Interpolated(
        ctypes.c_void_p(p_state), Energy_interp, ctypes.c_int(idx_chain)
    )
    return [E for E in Energy_interp]



"""
Minimum mode following (MMF)
-------------------------------------------------------------
"""

from spirit import spiritlib
from spirit.io import FILEFORMAT_OVF_TEXT
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

## ---------------------------------- Set ----------------------------------

_MMF_Set_Output_Tag = _spirit.Parameters_MMF_Set_Output_Tag
_MMF_Set_Output_Tag.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_Output_Tag.restype = None


def set_output_tag(p_state, tag, idx_image=-1, idx_chain=-1):
    """Set the tag placed in front of output file names.

    If the tag is "<time>", it will be the date-time of the creation of the state."""
    _MMF_Set_Output_Tag(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(tag.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_Output_Folder = _spirit.Parameters_MMF_Set_Output_Folder
_MMF_Set_Output_Folder.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_Output_Folder.restype = None


def set_output_folder(p_state, folder, idx_image=-1, idx_chain=-1):
    """Set the folder, where output files are placed."""
    _MMF_Set_Output_Folder(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(folder.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_Output_General = _spirit.Parameters_MMF_Set_Output_General
_MMF_Set_Output_General.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_Output_General.restype = None


def set_output_general(
    p_state, any=True, initial=False, final=False, idx_image=-1, idx_chain=-1
):
    """Set whether to write any output files at all."""
    _MMF_Set_Output_General(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(any),
        ctypes.c_bool(initial),
        ctypes.c_bool(final),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_Output_Energy = _spirit.Parameters_MMF_Set_Output_Energy
_MMF_Set_Output_Energy.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_Output_Energy.restype = None


def set_output_energy(
    p_state,
    step=False,
    archive=True,
    spin_resolved=False,
    divide_by_nos=True,
    add_readability_lines=True,
    idx_image=-1,
    idx_chain=-1,
):
    """Set whether to write energy output files.

    - `step`: whether to write a new file after each set of iterations
    - `archive`: whether to append to an archive file after each set of iterations
    - `spin_resolved`: whether to write a file containing the energy of each spin
    - `divide_by_nos`: whether to divide energies by the number of spins
    - `add_readability_lines`: whether to separate columns by lines
    """
    _MMF_Set_Output_Energy(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(step),
        ctypes.c_bool(archive),
        ctypes.c_bool(spin_resolved),
        ctypes.c_bool(divide_by_nos),
        ctypes.c_bool(add_readability_lines),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_Output_Configuration = _spirit.Parameters_MMF_Set_Output_Configuration
_MMF_Set_Output_Configuration.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_Output_Configuration.restype = None


def set_output_configuration(
    p_state,
    step=False,
    archive=True,
    filetype=FILEFORMAT_OVF_TEXT,
    idx_image=-1,
    idx_chain=-1,
):
    """Set whether to write spin configuration output files.

    - `step`: whether to write a new file after each set of iterations
    - `archive`: whether to append to an archive file after each set of iterations
    - `filetype`: the format in which the data is written
    """
    _MMF_Set_Output_Configuration(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(step),
        ctypes.c_bool(archive),
        ctypes.c_int(filetype),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_N_Iterations = _spirit.Parameters_MMF_Set_N_Iterations
_MMF_Set_N_Iterations.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_N_Iterations.restype = None


def set_iterations(p_state, n_iterations, n_iterations_log, idx_image=-1, idx_chain=-1):
    """Set the number of iterations and how often to log and write output.

    - `n_iterations`: the maximum number of iterations
    - `n_iterations_log`: the number of iterations after which status is logged and output written
    """
    _MMF_Set_N_Iterations(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_iterations),
        ctypes.c_int(n_iterations_log),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_N_Modes = _spirit.Parameters_MMF_Set_N_Modes
_MMF_Set_N_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_MMF_Set_N_Modes.restype = None


def set_n_modes(p_state, n_modes, idx_image=-1, idx_chain=-1):
    """Set the number of modes to be calculated at each iteration."""
    _MMF_Set_N_Modes(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_modes),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MMF_Set_N_Mode_Follow = _spirit.Parameters_MMF_Set_N_Mode_Follow
_MMF_Set_N_Mode_Follow.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Set_N_Mode_Follow.restype = None


def set_n_mode_follow(p_state, n_mode, idx_image=-1, idx_chain=-1):
    """Set the index of the mode to follow."""
    _MMF_Set_N_Mode_Follow(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_mode),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


## ---------------------------------- Get ----------------------------------

_MMF_Get_N_Iterations = _spirit.Parameters_MMF_Get_N_Iterations
_MMF_Get_N_Iterations.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_int,
    ctypes.c_int,
]
_MMF_Get_N_Iterations.restype = None


def get_iterations(p_state, idx_image=-1, idx_chain=-1):
    """Returns the maximum number of iterations and the step size."""
    n_iterations = ctypes.c_int()
    n_iterations_log = ctypes.c_int()
    _MMF_Get_N_Iterations(
        p_state,
        ctypes.pointer(n_iterations),
        ctypes.pointer(n_iterations_log),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return int(n_iterations.value), int(n_iterations_log.value)


_MMF_Get_N_Modes = _spirit.Parameters_MMF_Get_N_Modes
_MMF_Get_N_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_MMF_Get_N_Modes.restype = ctypes.c_int


def get_n_modes(p_state, idx_image=-1, idx_chain=-1):
    """Returns the number of modes calculated at each iteration."""
    return int(
        _MMF_Get_N_Modes(p_state, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    )


_MMF_Get_N_Mode_Follow = _spirit.Parameters_MMF_Get_N_Mode_Follow
_MMF_Get_N_Mode_Follow.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_MMF_Get_N_Mode_Follow.restype = ctypes.c_int


def get_n_mode_follow(p_state, idx_image=-1, idx_chain=-1):
    """Returns the index of the mode which to follow."""
    return int(
        _MMF_Get_N_Mode_Follow(
            p_state, ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )



version_major    = 2
version_minor    = 2
version_patch    = 0
version          = "2.2.0"
revision         = "e82250d3b1441"
version_full     = "2.2.0 (e82250d3b1441)"
compiler         = "GNU"
compiler_version = "11.3.0"
compiler_full    = "GNU (11.3.0)"
scalartype       = "double"
pinning          = "OFF"
defects          = "OFF"
cuda             = "OFF"
openmp           = "OFF"
threads          = "OFF"
fftw             = "OFF"


"""
Constants
====================
"""

from spirit import spiritlib
from spirit.scalar import scalar
import spirit.spiritlib as spiritlib
import ctypes

# Load Library
_spirit = spiritlib.load_spirit_library()

_mu_B = _spirit.Constants_mu_B
_mu_B.argtypes = None
_mu_B.restype = scalar
mu_B = _mu_B()
"""The Bohr Magneton [meV / T]"""

_mu_0 = _spirit.Constants_mu_0
_mu_0.argtypes = None
_mu_0.restype = scalar
mu_0 = _mu_0()
"""The vacuum permeability [T^2 m^3 / meV]"""

_k_B = _spirit.Constants_k_B
_k_B.argtypes = None
_k_B.restype = scalar
k_B = _k_B()
"""The Boltzmann constant [meV / K]"""

_hbar = _spirit.Constants_hbar
_hbar.argtypes = None
_hbar.restype = scalar
hbar = _hbar()
"""Planck's constant [meV*ps / rad]"""

_mRy = _spirit.Constants_mRy
_mRy.argtypes = None
_mRy.restype = scalar
mRy = _mRy()
"""Millirydberg [mRy / meV]"""

_gamma = _spirit.Constants_gamma
_gamma.argtypes = None
_gamma.restype = scalar
gamma = _gamma()
"""Gyromagnetic ratio of electron [rad / (ps*T)]"""

_g_e = _spirit.Constants_g_e
_g_e.argtypes = None
_g_e.restype = scalar
g_e = _g_e()
"""Electron g-factor [unitless]"""

_Pi = _spirit.Constants_Pi
_Pi.argtypes = None
_Pi.restype = scalar
pi = _Pi()
"""Pi [rad]"""



"""
Geometry
====================

Change or get info on the current geometrical configuration, e.g.
number of cells in the three crystal translation directions.
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### Imports
from spirit.scalar import scalar
from spirit import system

import numpy as np

### Bravais lattice types
BRAVAIS_LATTICE_IRREGULAR = 0
"""Irregular"""

BRAVAIS_LATTICE_RECTILINEAR = 1
"""Rectilinear"""

BRAVAIS_LATTICE_SC = 2
"""Simple cubic"""

BRAVAIS_LATTICE_HEX2D = 3
"""Hexagonal (60deg)"""

BRAVAIS_LATTICE_HEX2D_60 = 4
"""Hexagonal (60deg)"""

BRAVAIS_LATTICE_HEX2D_120 = 5
"""Hexagonal (120deg)"""

BRAVAIS_LATTICE_HCP = 6
"""Hexagonal close packed"""

BRAVAIS_LATTICE_BCC = 7
"""Body centered cubic"""

BRAVAIS_LATTICE_FCC = 8
"""Face centered cubic"""

### ---------------------------------- Set ----------------------------------

_Set_Bravais_Lattice_Type = _spirit.Geometry_Set_Bravais_Lattice_Type
_Set_Bravais_Lattice_Type.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Set_Bravais_Lattice_Type.restype = None


def set_bravais_lattice_type(p_state, lattice_type, idx_image=-1, idx_chain=-1):
    """Set the bravais vectors to one of the pre-defined lattice types.

    Note: `Irregular`, `Rectilinear` and `HCP` cannot be used."""
    _Set_Bravais_Lattice_Type(ctypes.c_void_p(p_state), ctypes.c_int(lattice_type))


_Set_N_Cells = _spirit.Geometry_Set_N_Cells
_Set_N_Cells.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int)]
_Set_N_Cells.restype = None


def set_n_cells(p_state, n_cells=[1, 1, 1], idx_image=-1, idx_chain=-1):
    """Set the number of basis cells along the three bravais vectors."""
    vec3 = ctypes.c_int * 3
    _Set_N_Cells(ctypes.c_void_p(p_state), vec3(*n_cells))


_Set_mu_s = _spirit.Geometry_Set_mu_s
_Set_mu_s.argtypes = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_Set_mu_s.restype = None


def set_mu_s(p_state, mu_s, idx_image=-1, idx_chain=-1):
    """Set the magnetic moment of all atoms."""
    _Set_mu_s(
        ctypes.c_void_p(p_state),
        ctypes.c_float(mu_s),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Cell_Atom_Types = _spirit.Geometry_Set_Cell_Atom_Types
_Set_Cell_Atom_Types.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_float),
]
_Set_Cell_Atom_Types.restype = None


def set_cell_atom_types(p_state, atom_types, idx_image=-1, idx_chain=-1):
    """Set the atom types of the atoms in the basis cell."""
    n = len(atom_types)
    vec = ctypes.c_int * n
    _Set_Cell_Atom_Types(ctypes.c_void_p(p_state), ctypes.c_int(n), vec(*atom_types))


_Set_Bravais_Vectors = _spirit.Geometry_Set_Bravais_Vectors
_Set_Bravais_Vectors.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
]
_Set_Bravais_Vectors.restype = None


def set_bravais_vectors(
    p_state,
    ta=[1.0, 0.0, 0.0],
    tb=[0.0, 1.0, 0.0],
    tc=[0.0, 0.0, 1.0],
    idx_image=-1,
    idx_chain=-1,
):
    """Manually specify the bravais vectors."""
    vec3 = ctypes.c_float * 3
    _Set_Bravais_Vectors(ctypes.c_void_p(p_state), vec3(ta), vec3(tb), vec3(tc))


_Set_Lattice_Constant = _spirit.Geometry_Set_Lattice_Constant
_Set_Lattice_Constant.argtypes = [ctypes.c_void_p, ctypes.c_float]
_Set_Lattice_Constant.restype = None


def set_lattice_constant(p_state, lattice_constant, idx_image=-1, idx_chain=-1):
    """Set the global lattice scaling constant."""
    _Set_Lattice_Constant(p_state, ctypes.c_float(lattice_constant))


### ---------------------------------- Get ----------------------------------

_Get_Bounds = _spirit.Geometry_Get_Bounds
_Get_Bounds.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Bounds.restype = None


def get_bounds(p_state, idx_image=-1, idx_chain=-1):
    """Get the bounds of the system in global coordinates.

    Returns two arrays of `shape(3)` containing minimum and maximum bounds respectively.
    """
    _min = (3 * ctypes.c_float)()
    _max = (3 * ctypes.c_float)()
    _Get_Bounds(
        ctypes.c_void_p(p_state),
        _min,
        _max,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return [_min[i] for i in range(3)], [_max[i] for i in range(3)]


_Get_Center = _spirit.Geometry_Get_Center
_Get_Center.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Center.restype = None


def get_center(p_state, idx_image=-1, idx_chain=-1):
    """Get the center of the system in global coordinates.

    Returns an array of `shape(3)`.
    """
    _center = (3 * ctypes.c_float)()
    _Get_Center(
        ctypes.c_void_p(p_state),
        _center,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return [_center[i] for i in range(3)]


_Get_Bravais_Lattice_Type = _spirit.Geometry_Get_Bravais_Lattice_Type
_Get_Bravais_Lattice_Type.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Bravais_Lattice_Type.restype = ctypes.c_int


def get_bravais_lattice_type(p_state, idx_image=-1, idx_chain=-1):
    """Get the bravais lattice type corresponding to one of the integers defined above."""
    return int(
        _Get_Bravais_Lattice_Type(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_Get_Bravais_Vectors = _spirit.Geometry_Get_Bravais_Vectors
_Get_Bravais_Vectors.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Bravais_Vectors.restype = None


def get_bravais_vectors(p_state, idx_image=-1, idx_chain=-1):
    """Get the Bravais vectors.

    Returns three arrays of `shape(3)`.
    """
    _a = (3 * ctypes.c_float)()
    _b = (3 * ctypes.c_float)()
    _c = (3 * ctypes.c_float)()
    _Get_Bravais_Vectors(
        ctypes.c_void_p(p_state),
        _a,
        _b,
        _c,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return [a for a in _a], [b for b in _b], [c for c in _c]


_Get_N_Cells = _spirit.Geometry_Get_N_Cells
_Get_N_Cells.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_N_Cells.restype = None


def get_n_cells(p_state, idx_image=-1, idx_chain=-1):
    """Get the number of basis cells along the three bravais vectors.

    Returns an array of `shape(3)`.
    """
    n_cells = (3 * ctypes.c_int)()
    _Get_N_Cells(
        ctypes.c_void_p(p_state),
        n_cells,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return [n for n in n_cells]


_Get_Dimensionality = _spirit.Geometry_Get_Dimensionality
_Get_Dimensionality.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Dimensionality.restype = ctypes.c_int


def get_dimensionality(p_state, idx_image=-1, idx_chain=-1):
    """Get the dimensionality of the geometry."""
    return int(
        _Get_Dimensionality(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


### Get Pointer to Spin Positions
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Positions = _spirit.Geometry_Get_Positions
_Get_Positions.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Positions.restype = ctypes.POINTER(scalar)


def get_positions(p_state, idx_image=-1, idx_chain=-1):
    """Returns a `numpy.array_view` of `shape(NOS, 3)` with the components of each spins position.

    Changing the contents of this array_view will have direct effect on the state and should not be done.
    """
    nos = system.get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar * 3 * nos
    Data = _Get_Positions(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = np.frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view


_Get_N_Cell_Atoms = _spirit.Geometry_Get_N_Cell_Atoms
_Get_N_Cell_Atoms.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_N_Cell_Atoms.restype = ctypes.c_int


def get_n_cell_atoms(p_state, idx_image=-1, idx_chain=-1):
    """Get the number of atoms in the basis cell."""
    return int(
        _Get_N_Cell_Atoms(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


### Get Pointer to atom types
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Atom_Types = _spirit.Geometry_Get_Atom_Types
_Get_Atom_Types.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Atom_Types.restype = ctypes.POINTER(ctypes.c_int)


def get_atom_types(p_state, idx_image=-1, idx_chain=-1):
    """Get the types of all atoms as a `numpy.array_view` of `shape(NOS)`.

    If e.g. disorder is activated, this allows to view and manipulate the types of individual atoms.
    """
    nos = system.get_nos(p_state, idx_image, idx_chain)
    ArrayType = ctypes.c_int * nos
    Data = _Get_Atom_Types(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = np.frombuffer(array_pointer.contents, dtype=ctypes.c_int)
    array_view = array.view()
    return array_view



"""
Landau-Lifshitz-Gilbert (LLG)
-------------------------------------------------------------
"""

from spirit import spiritlib
from spirit.io import FILEFORMAT_OVF_TEXT
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### ---------------------------------- Set ----------------------------------

_LLG_Set_Output_Tag = _spirit.Parameters_LLG_Set_Output_Tag
_LLG_Set_Output_Tag.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Output_Tag.restype = None


def set_output_tag(p_state, tag, idx_image=-1, idx_chain=-1):
    """Set the tag placed in front of output file names.

    If the tag is "<time>", it will be the date-time of the creation of the state."""
    _LLG_Set_Output_Tag(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(tag.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Output_Folder = _spirit.Parameters_LLG_Set_Output_Folder
_LLG_Set_Output_Folder.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Output_Folder.restype = None


def set_output_folder(p_state, folder, idx_image=-1, idx_chain=-1):
    """Set the folder, where output files are placed."""
    _LLG_Set_Output_Folder(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(folder.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Output_General = _spirit.Parameters_LLG_Set_Output_General
_LLG_Set_Output_General.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Output_General.restype = None


def set_output_general(
    p_state, any=True, initial=False, final=False, idx_image=-1, idx_chain=-1
):
    """Set whether to write any output files at all."""
    _LLG_Set_Output_General(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(any),
        ctypes.c_bool(initial),
        ctypes.c_bool(final),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Output_Energy = _spirit.Parameters_LLG_Set_Output_Energy
_LLG_Set_Output_Energy.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Output_Energy.restype = None


def set_output_energy(
    p_state,
    step=False,
    archive=True,
    spin_resolved=False,
    divide_by_nos=True,
    add_readability_lines=True,
    idx_image=-1,
    idx_chain=-1,
):
    """Set whether to write energy output files.

    - `step`: whether to write a new file after each set of iterations
    - `archive`: whether to append to an archive file after each set of iterations
    - `spin_resolved`: whether to write a file containing the energy of each spin
    - `divide_by_nos`: whether to divide energies by the number of spins
    - `add_readability_lines`: whether to separate columns by lines
    """
    _LLG_Set_Output_Energy(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(step),
        ctypes.c_bool(archive),
        ctypes.c_bool(spin_resolved),
        ctypes.c_bool(divide_by_nos),
        ctypes.c_bool(add_readability_lines),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Output_Configuration = _spirit.Parameters_LLG_Set_Output_Configuration
_LLG_Set_Output_Configuration.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Output_Configuration.restype = None


def set_output_configuration(
    p_state,
    step=False,
    archive=True,
    filetype=FILEFORMAT_OVF_TEXT,
    idx_image=-1,
    idx_chain=-1,
):
    """Set whether to write spin configuration output files.

    - `step`: whether to write a new file after each set of iterations
    - `archive`: whether to append to an archive file after each set of iterations
    - `filetype`: the format in which the data is written
    """
    _LLG_Set_Output_Configuration(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(step),
        ctypes.c_bool(archive),
        ctypes.c_int(filetype),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_N_Iterations = _spirit.Parameters_LLG_Set_N_Iterations
_LLG_Set_N_Iterations.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_N_Iterations.restype = None


def set_iterations(p_state, n_iterations, n_iterations_log, idx_image=-1, idx_chain=-1):
    """Set the number of iterations and how often to log and write output.

    - `n_iterations`: the maximum number of iterations
    - `n_iterations_log`: the number of iterations after which status is logged and output written
    """
    _LLG_Set_N_Iterations(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_iterations),
        ctypes.c_int(n_iterations_log),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Direct_Minimization = _spirit.Parameters_LLG_Set_Direct_Minimization
_LLG_Set_Direct_Minimization.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Direct_Minimization.restype = None


def set_direct_minimization(p_state, use_minimization, idx_image=-1, idx_chain=-1):
    """Set whether to minimise the energy without precession.

    This only influences dynamics solvers, which will then perform pseudodynamics,
    simulating only the damping part of the LLG equation.
    """
    _LLG_Set_Direct_Minimization(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(use_minimization),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Convergence = _spirit.Parameters_LLG_Set_Convergence
_LLG_Set_Convergence.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Convergence.restype = None


def set_convergence(p_state, convergence, idx_image=-1, idx_chain=-1):
    """Set the convergence limit.

    When the maximum absolute component value of the force drops below this value,
    the calculation is considered converged and will stop.
    """
    _LLG_Set_Convergence(
        ctypes.c_void_p(p_state),
        ctypes.c_float(convergence),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Time_Step = _spirit.Parameters_LLG_Set_Time_Step
_LLG_Set_Time_Step.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Time_Step.restype = None


def set_timestep(p_state, dt, idx_image=-1, idx_chain=-1):
    """Set the time step [ps] for the calculation."""
    _LLG_Set_Time_Step(p_state, ctypes.c_float(dt), idx_image, idx_chain)


_LLG_Set_Damping = _spirit.Parameters_LLG_Set_Damping
_LLG_Set_Damping.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Damping.restype = None


def set_damping(p_state, damping, idx_image=-1, idx_chain=-1):
    """Set the Gilbert damping parameter [unitless]."""
    _LLG_Set_Damping(
        ctypes.c_void_p(p_state),
        ctypes.c_float(damping),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_STT = _spirit.Parameters_LLG_Set_STT
_LLG_Set_STT.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_float,
    ctypes.POINTER(3 * ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_STT.restype = None


_LLG_Set_Non_Adiabatic_Damping = _spirit.Parameters_LLG_Set_Non_Adiabatic_Damping
_LLG_Set_Non_Adiabatic_Damping.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Non_Adiabatic_Damping.restype = None


def set_non_adiabatic_damping(p_state, beta, idx_image=-1, idx_chain=-1):
    """Set the Gilbert damping parameter [unitless]."""
    _LLG_Set_Non_Adiabatic_Damping(
        ctypes.c_void_p(p_state),
        ctypes.c_float(beta),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_STT = _spirit.Parameters_LLG_Set_STT
_LLG_Set_STT.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_float,
    ctypes.POINTER(3 * ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_STT.restype = None


def set_stt(p_state, use_gradient, magnitude, direction, idx_image=-1, idx_chain=-1):
    """Set the spin current configuration.

    - `use_gradient`: `True`: use the spatial gradient, `False`: monolayer approximation
    - `magnitude`: current strength
    - `direction`: current direction or polarisation direction, array of `shape(3)`
    """
    vec3 = ctypes.c_float * 3
    direction = vec3(*direction)
    ctypes.cast(direction, ctypes.POINTER(vec3))
    _LLG_Set_STT(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(use_gradient),
        ctypes.c_float(magnitude),
        direction,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_LLG_Set_Temperature = _spirit.Parameters_LLG_Set_Temperature
_LLG_Set_Temperature.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Temperature.restype = None
_LLG_Set_Temperature_Gradient = _spirit.Parameters_LLG_Set_Temperature_Gradient
_LLG_Set_Temperature_Gradient.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.POINTER(3 * ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Set_Temperature_Gradient.restype = None


def set_temperature(
    p_state,
    temperature,
    gradient_inclination=0,
    gradient_direction=[1.0, 0.0, 0.0],
    idx_image=-1,
    idx_chain=-1,
):
    """Set the temperature configuration.

    - `temperature`: homogeneous base temperature [K]
    - `gradient_inclination`: inclination of the temperature gradient [K/a]
    - `gradient_direction`: direction of the temperature gradient, array of `shape(3)`
    """
    _LLG_Set_Temperature(
        ctypes.c_void_p(p_state),
        ctypes.c_float(temperature),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    vec3 = ctypes.c_float * 3
    gradient_direction = vec3(*gradient_direction)
    ctypes.cast(gradient_direction, ctypes.POINTER(vec3))
    _LLG_Set_Temperature_Gradient(
        ctypes.c_void_p(p_state),
        ctypes.c_float(gradient_inclination),
        gradient_direction,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


### ---------------------------------- Get ----------------------------------

_LLG_Get_N_Iterations = _spirit.Parameters_LLG_Get_N_Iterations
_LLG_Get_N_Iterations.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Get_N_Iterations.restype = None


def get_iterations(p_state, idx_image=-1, idx_chain=-1):
    """Returns the maximum number of iterations and the step size."""
    n_iterations = ctypes.c_int()
    n_iterations_log = ctypes.c_int()
    _LLG_Get_N_Iterations(
        p_state,
        ctypes.pointer(n_iterations),
        ctypes.pointer(n_iterations_log),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return int(n_iterations.value), int(n_iterations_log.value)


_LLG_Get_Direct_Minimization = _spirit.Parameters_LLG_Get_Direct_Minimization
_LLG_Get_Direct_Minimization.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Direct_Minimization.restype = ctypes.c_bool


def get_direct_minimization(p_state, idx_image=-1, idx_chain=-1):
    """Returns whether only energy minimisation will be performed."""
    return float(
        _LLG_Get_Direct_Minimization(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_LLG_Get_Convergence = _spirit.Parameters_LLG_Get_Convergence
_LLG_Get_Convergence.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Convergence.restype = ctypes.c_float


def get_convergence(p_state, idx_image=-1, idx_chain=-1):
    """Returns the convergence value."""
    return float(
        _LLG_Get_Convergence(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_LLG_Get_Time_Step = _spirit.Parameters_LLG_Get_Time_Step
_LLG_Get_Time_Step.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Time_Step.restype = ctypes.c_float


def get_timestep(p_state, idx_image=-1, idx_chain=-1):
    """Returns the time step [ps]."""
    return float(
        _LLG_Get_Time_Step(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_LLG_Get_Damping = _spirit.Parameters_LLG_Get_Damping
_LLG_Get_Damping.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Damping.restype = ctypes.c_float


def get_damping(p_state, idx_image=-1, idx_chain=-1):
    """Returns the Gilbert damping parameter."""
    return float(
        _LLG_Get_Damping(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_LLG_Get_STT = _spirit.Parameters_LLG_Get_STT
_LLG_Get_STT.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_bool),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Get_STT.restype = None


_LLG_Get_Non_Adiabatic_Damping = _spirit.Parameters_LLG_Get_Non_Adiabatic_Damping
_LLG_Get_Non_Adiabatic_Damping.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Non_Adiabatic_Damping.restype = ctypes.c_float


def get_non_adiabatic_damping(p_state, idx_image=-1, idx_chain=-1):
    """Returns the non-adiabatic damping parameter."""
    return float(
        _LLG_Get_Non_Adiabatic_Damping(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_LLG_Get_STT = _spirit.Parameters_LLG_Get_STT
_LLG_Get_STT.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_bool),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Get_STT.restype = None


def get_stt(p_state, idx_image=-1, idx_chain=-1):
    """Returns the spin current configuration.

    - magnitude
    - direction, array of `shape(3)`
    - whether the spatial gradient is used
    """
    direction = (3 * ctypes.c_float)()
    use_gradient = ctypes.c_bool()
    magnitude = ctypes.c_float()
    _LLG_Get_STT(
        ctypes.c_void_p(p_state),
        ctypes.pointer(use_gradient),
        ctypes.pointer(magnitude),
        direction,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return float(magnitude.value), [float(i) for i in direction], bool(use_gradient)


_LLG_Get_Temperature = _spirit.Parameters_LLG_Get_Temperature
_LLG_Get_Temperature.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Temperature.restype = ctypes.c_float
_LLG_Get_Temperature_Gradient = _spirit.Parameters_LLG_Get_Temperature_Gradient
_LLG_Get_Temperature_Gradient.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_LLG_Get_Temperature_Gradient.restype = None


def get_temperature(p_state, idx_image=-1, idx_chain=-1):
    """Returns the temperature configuration.

    - global base temperature [K]
    - inclination of the temperature gradient [K/a]
    - direction of the temperature gradient, array of `shape(3)`
    """
    temperature = float(
        _LLG_Get_Temperature(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )
    gradient_inclination = ctypes.c_float()
    gradient_direction = (3 * ctypes.c_float)()
    _LLG_Get_Temperature_Gradient(
        ctypes.c_void_p(p_state),
        ctypes.pointer(gradient_inclination),
        gradient_direction,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return temperature, gradient_inclination, gradient_direction



"""
Configuration
====================

Set various spin configurations, such as homogeneous domains, spirals or skyrmions.

All configuration setters support the following arguments with default values:

- `pos=[0,0,0]`: the centre of the configuration, relative to the centre of the system
- `border_rectangular=[-1,-1,-1]`: values > 0 mean a restriction in `+` and `-` direction relative to the position
- `border_cylindrical=-1`: restricts the initialisation to a z-aligned cylinder around the position
- `border_spherical=-1`: restricts the initialisation to a sphere around the position
- `inverted=False`: exactly inverts the above restrictions
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

_Domain = _spirit.Configuration_Domain
_Domain.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Domain.restype = None


def domain(
    p_state,
    dir,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set a domain (homogeneous) configuration."""
    vec3 = ctypes.c_float * 3
    _Domain(
        ctypes.c_void_p(p_state),
        vec3(*dir),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_PlusZ = _spirit.Configuration_PlusZ
_PlusZ.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_PlusZ.restype = None


def plus_z(
    p_state,
    pos=[0.0, 0.0, 0.0],
    border_rectangular=[-1.0, -1.0, -1.0],
    border_cylindrical=-1.0,
    border_spherical=-1.0,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set a `+z` (homogeneous) configuration."""
    vec3 = ctypes.c_float * 3
    _PlusZ(
        ctypes.c_void_p(p_state),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_MinusZ = _spirit.Configuration_MinusZ
_MinusZ.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_MinusZ.restype = None


def minus_z(
    p_state,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set a `-z` (homogeneous) configuration."""
    vec3 = ctypes.c_float * 3
    _MinusZ(
        ctypes.c_void_p(p_state),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Random = _spirit.Configuration_Random
_Random.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Random.restype = None


def random(
    p_state,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Distribute all spins randomly on the unit sphere."""
    vec3 = ctypes.c_float * 3
    _Random(
        ctypes.c_void_p(p_state),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        False,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Add_Noise_Temperature = _spirit.Configuration_Add_Noise_Temperature
_Add_Noise_Temperature.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Add_Noise_Temperature.restype = None


def add_noise(
    p_state,
    temperature,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Add temperature-scaled random noise to configuration."""
    vec3 = ctypes.c_float * 3
    _Add_Noise_Temperature(
        ctypes.c_void_p(p_state),
        ctypes.c_float(temperature),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Skyrmion = _spirit.Configuration_Skyrmion
_Skyrmion.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Skyrmion.restype = None


def skyrmion(
    p_state,
    radius,
    order=1,
    phase=1,
    up_down=False,
    achiral=False,
    right_left=False,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set a skyrmion configuration.

    Arguments:

    - `radius`: the extent of the skyrmion, at which it points approximately upwards. The skyrmion only extends up to `radius`, meaning that `border_cylindrical` is not usually necessary.

    Keyword arguments:

    - `order`: the number of twists along a circle cutting the skyrmion
    - `phase`: 0 corresponds to a Neel skyrmion, -90 to a Bloch skyrmion
    - `up_down`: if `True`, the z-orientation is inverted
    - `achiral`: if `True`, the topological charge is inverted
    - `right_left`: if `True`, the in-plane rotation is inverted

    """
    vec3 = ctypes.c_float * 3
    _Skyrmion(
        ctypes.c_void_p(p_state),
        ctypes.c_float(radius),
        ctypes.c_float(order),
        ctypes.c_float(phase),
        ctypes.c_bool(up_down),
        ctypes.c_bool(achiral),
        ctypes.c_bool(right_left),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_DW_Skyrmion = _spirit.Configuration_DW_Skyrmion
_DW_Skyrmion.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_DW_Skyrmion.restype = None


def dw_skyrmion(
    p_state,
    dw_radius,
    dw_width,
    order=1,
    phase=1,
    up_down=False,
    achiral=False,
    right_left=False,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set a 360 degree domain wall skyrmion configuration.

    Arguments:

    - `dw_radius`: the radius of the circular domain wall skyrmion.
    - `dw_width`: the width of the domain wall circumference of the skyrmion.

    Keyword arguments:

    - `order`: the number of twists along a circle cutting the skyrmion
    - `phase`: 0 corresponds to a Neel skyrmion, -90 to a Bloch skyrmion
    - `up_down`: if `True`, the z-orientation is inverted
    - `achiral`: if `True`, the topological charge is inverted
    - `right_left`: if `True`, the in-plane rotation is inverted

    """
    vec3 = ctypes.c_float * 3
    _DW_Skyrmion(
        ctypes.c_void_p(p_state),
        ctypes.c_float(dw_radius),
        ctypes.c_float(dw_width),
        ctypes.c_float(order),
        ctypes.c_float(phase),
        ctypes.c_bool(up_down),
        ctypes.c_bool(achiral),
        ctypes.c_bool(right_left),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Hopfion = _spirit.Configuration_Hopfion
_Hopfion.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Hopfion.restype = None


def hopfion(
    p_state,
    radius,
    order=1,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    normal=[0, 0, 1],
    idx_image=-1,
    idx_chain=-1,
):
    """Set a Hopfion configuration.

    Arguments:

    - `radius`: the distance from the center to the center of the corresponding tubular isosurface

    Keyword arguments:

    - `order`: the number of windings of the toroidal hopfion

    In contrast to the skyrmion, it extends over the whole allowed space.
    """
    vec3 = ctypes.c_float * 3
    _Hopfion(
        ctypes.c_void_p(p_state),
        ctypes.c_float(radius),
        ctypes.c_int(order),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        vec3(*normal),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_SpinSpiral = _spirit.Configuration_SpinSpiral
_SpinSpiral.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_SpinSpiral.restype = None


def spin_spiral(
    p_state,
    direction_type,
    q_vector,
    axis,
    theta,
    pos=[0, 0, 0],
    border_rectangular=[-1, -1, -1],
    border_cylindrical=-1,
    border_spherical=-1,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set a spin spiral configuration.

    TODO: document parameters

    - `direction_type`:
    - `q_vector`:
    - `axis`:
    - `theta`:
    """
    vec3 = ctypes.c_float * 3
    _SpinSpiral(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(direction_type.encode("utf-8")),
        vec3(*q_vector),
        vec3(*axis),
        ctypes.c_float(theta),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Pinned = _spirit.Configuration_Set_Pinned
_Set_Pinned.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Pinned.restype = None


def set_pinned(
    p_state,
    pinned,
    pos=[0.0, 0.0, 0.0],
    border_rectangular=[-1.0, -1.0, -1.0],
    border_cylindrical=-1.0,
    border_spherical=-1.0,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set whether the spins within the given region are pinned or not.

    If they are pinned, they are pinned to their current orientation.
    """
    vec3 = ctypes.c_float * 3
    _Set_Pinned(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(pinned),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Atom_Type = _spirit.Configuration_Set_Atom_Type
_Set_Atom_Type.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Atom_Type.restype = None


def set_atom_type(
    p_state,
    atom_type=0,
    pos=[0.0, 0.0, 0.0],
    border_rectangular=[-1.0, -1.0, -1.0],
    border_cylindrical=-1.0,
    border_spherical=-1.0,
    inverted=False,
    idx_image=-1,
    idx_chain=-1,
):
    """Set the type of the atoms in the given region (default: 0).

    This can be used e.g. to insert defects (-1).
    """
    vec3 = ctypes.c_float * 3
    _Set_Atom_Type(
        ctypes.c_void_p(p_state),
        ctypes.c_int(atom_type),
        vec3(*pos),
        vec3(*border_rectangular),
        ctypes.c_float(border_cylindrical),
        ctypes.c_float(border_spherical),
        ctypes.c_bool(inverted),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )



"""
State
====================

The state contains the chain of spin systems (also called images).

To create a new state with containing a single image and run a simple example
of a **spin dynamics simulation**:

.. code-block:: python
    # Import spirit
    from spirit import state
    from spirit import simulation

    # Create a state using an input file
    with state.State("input/input.cfg") as p_state:
        # Start a LLG simulation using the SIB solver
        simulation.start(p_state, simulation.METHOD_LLG, simulation.SOLVER_SIB)

or call setup and delete manually:

.. code-block:: python
    # Import spirit
    from spirit import state
    from spirit import simulation

    # Create a state using an input file
    p_state = state.setup("input/input.cfg")
    # Start a LLG simulation using the SIB solver
    simulation.start(p_state, simulation.METHOD_LLG, simulation.SOLVER_SIB)
    # Delete the state and free memory etc.
    state.delete(p_state)

You can pass an input file specifying your initial system parameters.
If you do not pass an input file, the implemented defaults are used.
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()


### State wrapper class to be used in 'with' statement
class State:
    """Wrapper Class for a Spirit State.

    Can be used as `with spirit.state.State() as p_state:`
    """

    def __init__(self, configfile="", quiet=False):
        self.p_state = setup(configfile, quiet)

    def __enter__(self):
        return self.p_state

    def __exit__(self, exc_type, exc_value, traceback):
        delete(self.p_state)


### Setup State
_State_Setup = _spirit.State_Setup
_State_Setup.argtypes = [ctypes.c_char_p, ctypes.c_bool]
_State_Setup.restype = ctypes.c_void_p


def setup(configfile="", quiet=False):
    return _State_Setup(
        ctypes.c_char_p(configfile.encode("utf-8")), ctypes.c_bool(quiet)
    )


### Delete State
_State_Delete = _spirit.State_Delete
_State_Delete.argtypes = [ctypes.c_void_p]
_State_Delete.restype = None


def delete(p_state):
    return _State_Delete(ctypes.c_void_p(p_state))


### Get the date-time of the State creation
_State_To_Config = _spirit.State_To_Config
_State_To_Config.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]
_State_To_Config.restype = None


def to_config(p_state, filename, comment=""):
    """Write a config (input) file corresponding to the current parameters etc. of the state."""
    _State_To_Config(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_char_p(comment.encode("utf-8")),
    )


### Get the date-time of the State creation
_State_DateTime = _spirit.State_DateTime
_State_DateTime.argtypes = [ctypes.c_void_p]
_State_DateTime.restype = ctypes.c_char_p


def date_time(p_state):
    """Returns a string containing the date-time of the creation of the state."""
    return str(_State_DateTime(ctypes.c_void_p(p_state)))



"""
Hamiltonian
====================

Set the parameters of the Heisenberg Hamiltonian, such as external field or exchange interaction.
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### DM vector chirality
CHIRALITY_BLOCH = 1
"""DMI Bloch chirality type for neighbour shells"""

CHIRALITY_NEEL = 2
"""DMI Neel chirality type for neighbour shells"""

CHIRALITY_BLOCH_INVERSE = -1
"""DMI Bloch chirality type for neighbour shells with opposite sign"""

CHIRALITY_NEEL_INVERSE = -2
"""DMI Neel chirality type for neighbour shells with opposite sign"""

### DDI METHOD
DDI_METHOD_NONE = 0
"""Dipole-dipole interaction: do not calculate"""

DDI_METHOD_FFT = 1
"""Dipole-dipole interaction: use FFT convolutions"""

DDI_METHOD_FMM = 2
"""Dipole-dipole interaction: use a fast multipole method (FMM)"""

DDI_METHOD_CUTOFF = 3
"""Dipole-dipole interaction: use a direct summation with a cutoff radius"""

### ---------------------------------- Set ----------------------------------

_Set_Boundary_Conditions = _spirit.Hamiltonian_Set_Boundary_Conditions
_Set_Boundary_Conditions.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_bool),
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Boundary_Conditions.restype = None


def set_boundary_conditions(p_state, boundaries, idx_image=-1, idx_chain=-1):
    """Set the boundary conditions along the translation directions [a, b, c].

    0 = open, 1 = periodical
    """
    bool3 = ctypes.c_bool * 3
    _Set_Boundary_Conditions(
        ctypes.c_void_p(p_state),
        bool3(*boundaries),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Field = _spirit.Hamiltonian_Set_Field
_Set_Field.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Field.restype = None


def set_field(p_state, magnitude, direction, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) external magnetic field."""
    vec3 = ctypes.c_float * 3
    _Set_Field(
        ctypes.c_void_p(p_state),
        ctypes.c_float(magnitude),
        vec3(*direction),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Anisotropy = _spirit.Hamiltonian_Set_Anisotropy
_Set_Anisotropy.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Anisotropy.restype = None


def set_anisotropy(p_state, magnitude, direction, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) magnetocrystalline anisotropy."""
    vec3 = ctypes.c_float * 3
    _Set_Anisotropy(
        ctypes.c_void_p(p_state),
        ctypes.c_float(magnitude),
        vec3(*direction),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Exchange = _spirit.Hamiltonian_Set_Exchange
_Set_Exchange.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Exchange.restype = None


_Set_Cubic_Anisotropy = _spirit.Hamiltonian_Set_Cubic_Anisotropy
_Set_Cubic_Anisotropy.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Cubic_Anisotropy.restype = None


def set_cubic_anisotropy(p_state, magnitude, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) magnetocrystalline anisotropy."""
    _Set_Cubic_Anisotropy(
        ctypes.c_void_p(p_state),
        ctypes.c_float(magnitude),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_Exchange = _spirit.Hamiltonian_Set_Exchange
_Set_Exchange.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Set_Exchange.restype = None


def set_exchange(p_state, n_shells, J_ij, idx_image=-1, idx_chain=-1):
    """Set the Exchange interaction in terms of neighbour shells."""
    vec = ctypes.c_float * n_shells
    _Set_Exchange(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_shells),
        vec(*J_ij),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_DMI = _spirit.Hamiltonian_Set_DMI
_Set_DMI.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_Set_DMI.restype = None


def set_dmi(
    p_state, n_shells, D_ij, chirality=CHIRALITY_BLOCH, idx_image=-1, idx_chain=-1
):
    """Set the Dzyaloshinskii-Moriya interaction in terms of neighbour shells."""
    vec = ctypes.c_float * n_shells
    _Set_DMI(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_shells),
        vec(*D_ij),
        ctypes.c_int(chirality),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Set_DDI = _spirit.Hamiltonian_Set_DDI
_Set_DDI.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_float,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Set_DDI.restype = None


def set_ddi(
    p_state,
    ddi_method,
    n_periodic_images=[4, 4, 4],
    radius=0.0,
    pb_zero_padding=True,
    idx_image=-1,
    idx_chain=-1,
):
    """Set the dipolar interaction calculation method.

    - `ddi_method`: one of the integers defined above
    - `n_periodic_images`: the number of periodical images in the three translation directions,
      taken into account when boundaries in the corresponding direction are periodical
    - `radius`: the cutoff radius for the direct summation method
    - `pb_zero_padding`: if `True` zero padding is used for periodical directions
    """
    vec3 = ctypes.c_int * 3
    _Set_DDI(
        ctypes.c_void_p(p_state),
        ctypes.c_int(ddi_method),
        vec3(*n_periodic_images),
        ctypes.c_float(radius),
        ctypes.c_bool(pb_zero_padding),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


### ---------------------------------- Get ----------------------------------

_Get_Name = _spirit.Hamiltonian_Get_Name
_Get_Name.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Name.restype = ctypes.c_char_p


def get_name(p_state, idx_image=-1, idx_chain=-1):
    """Returns a string containing the name of the Hamiltonian currently in use."""
    return str(
        _Get_Name(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_Get_Boundary_Conditions = _spirit.Hamiltonian_Get_Boundary_Conditions
_Get_Boundary_Conditions.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_bool),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Boundary_Conditions.restype = None


def get_boundary_conditions(p_state, idx_image=-1, idx_chain=-1):
    """Returns an array of `shape(3)` containing the boundary conditions in the
    three translation directions `[a, b, c]` of the lattice.
    """
    boundaries = (3 * ctypes.c_bool)()
    _Get_Boundary_Conditions(
        ctypes.c_void_p(p_state),
        boundaries,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return [bc for bc in boundaries]


_Get_Field = _spirit.Hamiltonian_Get_Field
_Get_Field.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Field.restype = None


def get_field(p_state, idx_image=-1, idx_chain=-1):
    """Returns the magnitude and an array of `shape(3)` containing the direction of
    the external magnetic field.
    """
    magnitude = ctypes.c_float()
    normal = (3 * ctypes.c_float)()
    _Get_Field(
        ctypes.c_void_p(p_state),
        ctypes.byref(magnitude),
        normal,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return float(magnitude.value), [n for n in normal]


_Get_DDI = _spirit.Hamiltonian_Get_DDI
_Get_DDI.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_bool),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_DDI.restype = None


def get_ddi(p_state, idx_image=-1, idx_chain=-1):
    """Returns a dictionary, containing information about the used ddi settings"""
    ddi_method = ctypes.c_int()
    n_periodic_images = (3 * ctypes.c_int)()
    cutoff_radius = ctypes.c_float()
    pb_zero_padding = ctypes.c_bool()
    _Get_DDI(
        ctypes.c_void_p(p_state),
        ctypes.byref(ddi_method),
        n_periodic_images,
        ctypes.byref(cutoff_radius),
        ctypes.byref(pb_zero_padding),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return {
        "ddi_method": ddi_method.value,
        "n_periodic_images": [i for i in n_periodic_images],
        "cutoff_radius": cutoff_radius.value,
        "pb_zero_padding": pb_zero_padding.value,
    }


_Write_Hessian = _spirit.Hamiltonian_Write_Hessian
_Write_Hessian.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Write_Hessian.restype = None


def write_hessian(p_state, filename, triplet_format=True, idx_image=-1, idx_chain=-1):
    """RWrites the embedding Hessian to a file"""
    _Write_Hessian(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(filename.encode("utf-8")),
        ctypes.c_bool(triplet_format),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Get_Anisotropy = _spirit.Hamiltonian_Get_Anisotropy
_Get_Anisotropy.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Anisotropy.restype = None


def get_anisotropy(p_state, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) magnetocrystalline anisotropy."""
    magnitude = ctypes.c_float()
    normal = (3 * ctypes.c_float)()
    _Get_Anisotropy(
        ctypes.c_void_p(p_state),
        ctypes.byref(magnitude),
        normal,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return float(magnitude.value), [n for n in normal]


_Get_Cubic_Anisotropy = _spirit.Hamiltonian_Get_Cubic_Anisotropy
_Get_Cubic_Anisotropy.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Cubic_Anisotropy.restype = None


def get_cubic_anisotropy(p_state, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) magnetocrystalline anisotropy."""
    magnitude = ctypes.c_float()
    _Get_Cubic_Anisotropy(
        ctypes.c_void_p(p_state),
        ctypes.byref(magnitude),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return float(magnitude.value)



"""
Geodesic nudged elastic band (GNEB)
-------------------------------------------------------------
"""

from spirit import spiritlib
from spirit.io import FILEFORMAT_OVF_TEXT
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

# GNEB image types
IMAGE_NORMAL = 0
"""Regular GNEB image type."""

IMAGE_CLIMBING = 1
"""Climbing GNEB image type.
Climbing images move towards maxima along the path.
"""

IMAGE_FALLING = 2
"""Falling GNEB image type.
Falling images move towards the closest minima.
"""

IMAGE_STATIONARY = 3
"""Stationary GNEB image type.
Stationary images are not influenced during a GNEB calculation.
"""

### ---------------------------------- Set ----------------------------------

_GNEB_Set_Output_Tag = _spirit.Parameters_GNEB_Set_Output_Tag
_GNEB_Set_Output_Tag.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Output_Tag.restype = None


def set_output_tag(p_state, tag, idx_image=-1, idx_chain=-1):
    """Set the tag placed in front of output file names.

    If the tag is "<time>", it will be the date-time of the creation of the state."""
    _GNEB_Set_Output_Tag(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(tag.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Output_Folder = _spirit.Parameters_GNEB_Set_Output_Folder
_GNEB_Set_Output_Folder.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Output_Folder.restype = None


def set_output_folder(p_state, folder, idx_image=-1, idx_chain=-1):
    """Set the folder, where output files are placed."""
    _GNEB_Set_Output_Folder(
        ctypes.c_void_p(p_state),
        ctypes.c_char_p(folder.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Output_General = _spirit.Parameters_GNEB_Set_Output_General
_GNEB_Set_Output_General.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Output_General.restype = None


def set_output_general(
    p_state, any=True, initial=False, final=False, idx_image=-1, idx_chain=-1
):
    """Set whether to write any output files at all."""
    _GNEB_Set_Output_General(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(any),
        ctypes.c_bool(initial),
        ctypes.c_bool(final),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Output_Energies = _spirit.Parameters_GNEB_Set_Output_Energies
_GNEB_Set_Output_Energies.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Output_Energies.restype = None


def set_output_energies(
    p_state,
    step=True,
    interpolated=True,
    divide_by_nos=True,
    add_readability_lines=True,
    idx_image=-1,
    idx_chain=-1,
):
    """Set whether to write energy output files.

    - `step`: whether to write a new file after each set of iterations
    - `interpolated`: whether to write a file containing interpolated reaction coordinate and energy values
    - `divide_by_nos`: whether to divide energies by the number of spins
    - `add_readability_lines`: whether to separate columns by lines
    """
    _GNEB_Set_Output_Energies(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(step),
        ctypes.c_bool(interpolated),
        ctypes.c_bool(divide_by_nos),
        ctypes.c_bool(add_readability_lines),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Output_Chain = _spirit.Parameters_GNEB_Set_Output_Chain
_GNEB_Set_Output_Chain.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Output_Chain.restype = None


def set_output_chain(
    p_state, step=False, filetype=FILEFORMAT_OVF_TEXT, idx_image=-1, idx_chain=-1
):
    """Set whether to write chain output files.

    - `step`: whether to write a new file after each set of iterations
    - `filetype`: the format in which the data is written
    """
    _GNEB_Set_Output_Chain(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(step),
        ctypes.c_int(filetype),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_N_Iterations = _spirit.Parameters_GNEB_Set_N_Iterations
_GNEB_Set_N_Iterations.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_N_Iterations.restype = None


def set_iterations(p_state, n_iterations, n_iterations_log, idx_image=-1, idx_chain=-1):
    """Set the number of iterations and how often to log and write output.

    - `n_iterations`: the maximum number of iterations
    - `n_iterations_log`: the number of iterations after which status is logged and output written
    """
    _GNEB_Set_N_Iterations(
        ctypes.c_void_p(p_state),
        ctypes.c_int(n_iterations),
        ctypes.c_int(n_iterations_log),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Convergence = _spirit.Parameters_GNEB_Set_Convergence
_GNEB_Set_Convergence.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Convergence.restype = None


def set_convergence(p_state, convergence, idx_image=-1, idx_chain=-1):
    """Set the convergence limit.

    When the maximum absolute component value of the force drops below this value,
    the calculation is considered converged and will stop.
    """
    _GNEB_Set_Convergence(
        ctypes.c_void_p(p_state),
        ctypes.c_float(convergence),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Spring_Constant = _spirit.Parameters_GNEB_Set_Spring_Constant
_GNEB_Set_Spring_Constant.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Spring_Constant.restype = None
_GNEB_Set_Spring_Force_Ratio = _spirit.Parameters_GNEB_Set_Spring_Force_Ratio
_GNEB_Set_Spring_Force_Ratio.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Spring_Force_Ratio.restype = None


def set_spring_force(p_state, spring_constant=1, ratio=0, idx_image=-1, idx_chain=-1):
    """Set the spring force constant and the ratio between energy and reaction coordinate."""
    _GNEB_Set_Spring_Constant(
        ctypes.c_void_p(p_state),
        ctypes.c_float(spring_constant),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    _GNEB_Set_Spring_Force_Ratio(
        ctypes.c_void_p(p_state),
        ctypes.c_float(ratio),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Path_Shortening_Constant = (
    _spirit.Parameters_GNEB_Set_Path_Shortening_Constant
)
_GNEB_Set_Path_Shortening_Constant.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Path_Shortening_Constant.restype = None


def set_path_shortening_constant(
    p_state, shortening_constant, idx_image=-1, idx_chain=-1
):
    """Set the path shortening constant."""
    _GNEB_Set_Path_Shortening_Constant(
        ctypes.c_void_p(p_state),
        ctypes.c_float(shortening_constant),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Moving_Endpoints = _spirit.Parameters_GNEB_Set_Moving_Endpoints
_GNEB_Set_Moving_Endpoints.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int]
_GNEB_Set_Moving_Endpoints.restype = None


def set_moving_endpoints(p_state, moving_endpoints, idx_chain=-1):
    """Set if moving endpoints should be used."""
    _GNEB_Set_Moving_Endpoints(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(moving_endpoints),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Translating_Endpoints = _spirit.Parameters_GNEB_Set_Translating_Endpoints
_GNEB_Set_Translating_Endpoints.argtypes = [
    ctypes.c_void_p,
    ctypes.c_bool,
    ctypes.c_int,
]
_GNEB_Set_Translating_Endpoints.restype = None


def set_translating_endpoints(p_state, translating_endpoints, idx_chain=-1):
    """Set if attracting endpoints should be used."""
    _GNEB_Set_Translating_Endpoints(
        ctypes.c_void_p(p_state),
        ctypes.c_bool(translating_endpoints),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Equilibrium_Delta_Rx = _spirit.Parameters_GNEB_Set_Equilibrium_Delta_Rx
_GNEB_Set_Equilibrium_Delta_Rx.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_float,
    ctypes.c_int,
]
_GNEB_Set_Equilibrium_Delta_Rx.restype = None


def set_equilibrium_delta_Rx(p_state, delta_Rx_left, delta_Rx_right, idx_chain=-1):
    """Set if moving endpoints should be used."""
    _GNEB_Set_Equilibrium_Delta_Rx(
        ctypes.c_void_p(p_state),
        ctypes.c_float(delta_Rx_left),
        ctypes.c_float(delta_Rx_right),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Climbing_Falling = _spirit.Parameters_GNEB_Set_Climbing_Falling
_GNEB_Set_Climbing_Falling.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Set_Climbing_Falling.restype = None


def set_climbing_falling(p_state, image_type, idx_image=-1, idx_chain=-1):
    """Set the GNEB image type (see the integers defined above)."""
    _GNEB_Set_Climbing_Falling(
        ctypes.c_void_p(p_state),
        ctypes.c_int(image_type),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_GNEB_Set_Image_Type_Automatically = (
    _spirit.Parameters_GNEB_Set_Image_Type_Automatically
)
_GNEB_Set_Image_Type_Automatically.argtypes = [ctypes.c_void_p, ctypes.c_int]
_GNEB_Set_Image_Type_Automatically.restype = None


def set_image_type_automatically(p_state, idx_chain=-1):
    """Automatically set GNEB image types.

    Minima along the path will be set to falling, maxima to climbing and the rest to regular.
    """
    _GNEB_Set_Image_Type_Automatically(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)
    )


### ---------------------------------- Get ----------------------------------

_GNEB_Get_N_Iterations = _spirit.Parameters_GNEB_Get_N_Iterations
_GNEB_Get_N_Iterations.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_int,
]
_GNEB_Get_N_Iterations.restype = None


def get_iterations(p_state, idx_image=-1, idx_chain=-1):
    """Returns the maximum number of iterations and the step size."""
    n_iterations = ctypes.c_int()
    n_iterations_log = ctypes.c_int()
    _GNEB_Get_N_Iterations(
        ctypes.c_void_p(p_state),
        ctypes.pointer(n_iterations),
        ctypes.pointer(n_iterations_log),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return int(n_iterations.value), int(n_iterations_log.value)


_GNEB_Get_Convergence = _spirit.Parameters_GNEB_Get_Convergence
_GNEB_Get_Convergence.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Convergence.restype = ctypes.c_float


def get_convergence(p_state, idx_image=-1, idx_chain=-1):
    """Returns the convergence value."""
    return float(
        _GNEB_Get_Convergence(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_GNEB_Get_Spring_Constant = _spirit.Parameters_GNEB_Get_Spring_Constant
_GNEB_Get_Spring_Constant.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Spring_Constant.restype = ctypes.c_float
_GNEB_Get_Spring_Force_Ratio = _spirit.Parameters_GNEB_Get_Spring_Force_Ratio
_GNEB_Get_Spring_Force_Ratio.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Spring_Force_Ratio.restype = ctypes.c_float


def get_spring_force(p_state, idx_image=-1, idx_chain=-1):
    """Returns the spring force constant and Ratio of energy to reaction coordinate."""
    constant = float(
        _GNEB_Get_Spring_Constant(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )
    ratio = float(
        _GNEB_Get_Spring_Force_Ratio(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )
    return constant, ratio


_GNEB_Get_Path_Shortening_Constant = (
    _spirit.Parameters_GNEB_Get_Path_Shortening_Constant
)
_GNEB_Get_Path_Shortening_Constant.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
]
_GNEB_Get_Path_Shortening_Constant.restype = ctypes.c_float


def get_path_shortening_constant(p_state, idx_image=-1, idx_chain=-1):
    """Return the path shortening constant."""
    return float(
        _GNEB_Get_Path_Shortening_Constant(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_GNEB_Get_Moving_Endpoints = _spirit.Parameters_GNEB_Get_Moving_Endpoints
_GNEB_Get_Moving_Endpoints.argtypes = [ctypes.c_void_p, ctypes.c_int]
_GNEB_Get_Moving_Endpoints.restype = ctypes.c_bool


def get_moving_endpoints(p_state, idx_chain=-1):
    """Return if moving endpoints are used."""
    return bool(
        _GNEB_Get_Moving_Endpoints(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain))
    )


_GNEB_Get_Translating_Endpoints = _spirit.Parameters_GNEB_Get_Translating_Endpoints
_GNEB_Get_Translating_Endpoints.argtypes = [ctypes.c_void_p, ctypes.c_int]
_GNEB_Get_Translating_Endpoints.restype = ctypes.c_bool


def get_translating_endpoints(p_state, idx_chain=-1):
    """Return if translating endpoints are used."""
    return bool(
        _GNEB_Get_Translating_Endpoints(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)
        )
    )


_GNEB_Get_Equilibrium_Delta_Rx = _spirit.Parameters_GNEB_Get_Equilibrium_Delta_Rx
_GNEB_Get_Equilibrium_Delta_Rx.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_GNEB_Get_Equilibrium_Delta_Rx.restype = None


def get_equilibrium_delta_Rx(p_state, idx_chain=-1):
    """Return the equilibrium delta_Rx for the moving endpoints."""
    delta_Rx_left = ctypes.c_float()
    delta_Rx_right = ctypes.c_float()

    _GNEB_Get_Equilibrium_Delta_Rx(
        ctypes.c_void_p(p_state),
        ctypes.byref(delta_Rx_left),
        ctypes.byref(delta_Rx_right),
        ctypes.c_int(idx_chain),
    )

    return [float(delta_Rx_left.value), float(delta_Rx_right.value)]


_GNEB_Get_Climbing_Falling = _spirit.Parameters_GNEB_Get_Climbing_Falling
_GNEB_Get_Climbing_Falling.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Climbing_Falling.restype = ctypes.c_int


def get_climbing_falling(p_state, idx_image=-1, idx_chain=-1):
    """Returns the integer of whether an image is regular, climbing, falling, or stationary.

    The integers are defined above."""
    return int(
        _GNEB_Get_Climbing_Falling(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


_GNEB_Get_N_Energy_Interpolations = _spirit.Parameters_GNEB_Get_N_Energy_Interpolations
_GNEB_Get_N_Energy_Interpolations.argtypes = [ctypes.c_void_p, ctypes.c_int]
_GNEB_Get_N_Energy_Interpolations.restype = ctypes.c_int


def get_n_energy_interpolations(p_state, idx_chain=-1):
    """Returns the number of energy values interpolated between images."""
    return int(
        _GNEB_Get_N_Energy_Interpolations(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)
        )
    )



"""
Transition
====================
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

_Homogeneous = _spirit.Transition_Homogeneous
_Homogeneous.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_Homogeneous.restype = None


def homogeneous(p_state, idx_1, idx_2, idx_chain=-1):
    """Generate homogeneous transition between two images of a chain."""
    _Homogeneous(
        ctypes.c_void_p(p_state),
        ctypes.c_int(idx_1),
        ctypes.c_int(idx_2),
        ctypes.c_int(idx_chain),
    )


_Homogeneous_Interpolate = _spirit.Transition_Homogeneous_Insert_Interpolated
_Homogeneous_Interpolate.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Homogeneous_Interpolate.restype = None


def homogeneous_insert_interpolated(p_state, n_interpolate, idx_chain=-1):
    """Make chain denser by inserting n_interpolate images between all images."""
    _Homogeneous_Interpolate(
        ctypes.c_void_p(p_state), ctypes.c_int(n_interpolate), ctypes.c_int(idx_chain)
    )


_Add_Noise_Temperature = _spirit.Transition_Add_Noise_Temperature
_Add_Noise_Temperature.argtypes = [
    ctypes.c_void_p,
    ctypes.c_float,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
]
_Add_Noise_Temperature.restype = None


def add_noise(p_state, temperature, idx_1, idx_2, idx_chain=-1):
    """Add some temperature-scaled noise to a transition between two images of a chain."""
    _Add_Noise_Temperature(
        ctypes.c_void_p(p_state),
        ctypes.c_float(temperature),
        ctypes.c_int(idx_1),
        ctypes.c_int(idx_2),
        ctypes.c_int(idx_chain),
    )
